/*
 * SPDX-FileCopyrightText: 2021-2022 Megan Conkle <megan.conkle@kdemail.net>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

::-webkit-scrollbar {
    width: 6px;  /* for vertical scrollbars */
    height: 6px; /* for horizontal scrollbars */
    background-color: transparent;
}

::-webkit-scrollbar-track {
    border-radius: $scrollbar-border-radius;
    background-color: transparent;
}

::-webkit-scrollbar-track:hover {
    background-color: $tertiary-fill-color;
}

::-webkit-scrollbar-thumb {
    min-width: 50px;
    min-height: 50px;
    background-color: $secondary-fill-color;
    border-radius: $scrollbar-border-radius;
}

::-webkit-scrollbar-thumb:hover {
    background-color: $accent-color;
}

::-webkit-scrollbar-button {
    width: 0;
    height: 0;
    display: none;
}
::-webkit-scrollbar-corner {
    background-color: transparent;
}

#livepreviewplaceholder {
    display: inline-block;
    position: fixed;
    overflow-x: scroll;
    overflow-y: scroll;
    width: -webkit-calc(100% - 16px);
    height: inherit;
    left: 0px;
    top: 8px;
    bottom: 0px;
    background: $background-color;
    margin: 0;
    padding: 5px;
}

body {
    margin: 0;
    padding: 0;
    overflow: hidden;
    color: $text-color;
    background-color: $background-color;
    font-family: "$body-font-family", ui-serif, serif;
    font-size: $body-font-size;
    line-height: 1.5;
}

h1 {
    color: $heading-color;
    font-weight: bold;
    text-transform: uppercase;
}

h2, h3, h4, h5, h6 {
    color: $heading-color;
    font-weight: bold;
    text-transform: none;
}

table {
    border: 1px solid $grid-color;
    text-align: left;
    border-collapse: collapse;
    width: 100%;
    line-height: normal;
}

tr {
    border: 1px solid $grid-color;
    text-align: left;
    margin: 0;
    padding: 10px;
}

tr:nth-child(even) {
    background-color: $tertiary-fill-color;
} 

th {
    font-weight: bold;
    text-align: left;
    border: 1px solid $grid-color;
    padding: 2px 5px 2px 5px;
}

td {
    border: 1px solid $grid-color;
    padding: 2px 5px 2px 5px;
}

pre {
    color: $code-color;
    background-color: transparent;
    font-family: "$code-font-family", monospace;
    border: 1px solid $separator-color;
    padding: 10px;
    border-radius: 3px;
    line-height: normal;
    font-size: $code-font-size;
    overflow: auto;
}

code {
    color: $code-color;
    font-family: "$code-font-family", monospace;
    font-size: $code-font-size;
}

a {
    color: $link-color;
}

img {
    max-width: 100%;
}

pre code, pre tt {
    background-color: transparent;
    border: none;
}

blockquote {
    color: $block-quote-color;
    background-color: transparent;
    border-left: 4px solid $separator-color;
    padding: 0 15px;
}

hr {
    color: $separator-color;
    background-color: $separator-color;
    border: 1px none;
    height: 1px;
}

ul, ol {
    line-height: 2;
}

kbd {
    color: $code-color;
    background-color: transparent;
    font-family: "$code-font-family", monospace;
    border: 1px solid $separator-color;
    border-radius: 3px;
    box-shadow: inset 0 -1px 0 $separator-color;
}
