import {Loader} from '../../../js/components/loader.js';

Loader.preLoad(
  'loader', 'startup',
  'core',
  'input/tex-full',
  'output/chtml', 'output/chtml/fonts/tex.js',
  'ui/menu', 'a11y/assistive-mml'
);
